﻿/*
 *
 *
 *    MainForm.cs
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using System.Xml;
using System.Text.RegularExpressions;

namespace KK5JY.RadioStudy.GROL {
	public partial class MainForm : Form {
		const string NoFileName = "(no file)";

		#region Inner Types
		private class Answer {
			public string Text;
			public char Letter;
		}

		private class Question {
			public string Number;
			public string Text;
			public List<Answer> Answers;
			public char Letter;

			public Question() {
				Answers = new List<Answer>(4);
			}
		}

		private enum States {
			None,
			Question,
			Answer,
		}
		#endregion

		#region Private Fields
		private string ConfigPath { get { return System.IO.Path.Combine(System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "RadioStudyGROL"); } }
		private string ConfigFile { get { return System.IO.Path.Combine(ConfigPath, "RadioStudyGROL.xml"); } }

		private string LastFile {
			get {
				return m_LastFile;
			}
			set {
				if (value == null) tsFileName.Text = NoFileName;
				else tsFileName.Text = System.IO.Path.GetFileName(value);
				m_LastFile = value;
			}
		}

		private Dictionary<string, Question> m_Questions;
		private string m_LastPath;
		private string m_LastFile;
		private int m_SelectedIndex;
		private Question m_SelectedQuestion;
		private bool m_Capture;
		private Image m_Yes;
		private Image m_No;
		#endregion

		public MainForm() {
			InitializeComponent();
			m_Questions = new Dictionary<string, Question>();
			var me = System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
			m_LastPath = System.IO.Path.Combine(me, "Documents", "FCC", "GROL");
			m_Yes = Image.FromFile(Path.Combine(me, "yes.png"));
			m_No = Image.FromFile(Path.Combine(me, "no.png"));
			bA.Tag = 'A';
			bB.Tag = 'B';
			bC.Tag = 'C';
			bD.Tag = 'D';
		}

		#region Event Handlers
		protected override void OnClosing(CancelEventArgs e) {
			SaveConfig();
			base.OnClosing(e);
		}

		protected override void OnShown(EventArgs e) {
			LoadConfig();
			base.OnShown(e);
		}

		private void OpenClick(object sender, EventArgs e) {
			using (OpenFileDialog dialog = new OpenFileDialog {
				AddExtension = true,
				AutoUpgradeEnabled = true,
				CheckFileExists = true,
				DefaultExt = "txt",
				Filter = "Text Files|*.txt",
				InitialDirectory = m_LastPath,
				Multiselect = false,
				Title = "Open a new exam text" }) {
				if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
					var file = dialog.FileName;
					ProcessFile(file);
				}
			}
		}

		private void AnswerClick(object sender, EventArgs e) {
			if (m_Questions.Count == 0) return;
			if (ReferenceEquals(imgResult.Image, m_Yes)) {
				NextClick(sender, e);
				return;
			}

			imgResult.Image = null;
			var answer = (char)((sender as Button).Tag);
			ProcessAnswer(answer);
		}

		private void PrevClick(object sender, EventArgs e) {
			if (m_Questions.Count == 0) return;
			if (m_SelectedIndex == 0) m_SelectedIndex = m_Questions.Count - 1;
			else m_SelectedIndex--;
			LoadQuestion(m_SelectedIndex);
		}

		private void NextClick(object sender, EventArgs e) {
			if (m_Questions.Count == 0) return;
			m_SelectedIndex = (m_SelectedIndex + 1) % m_Questions.Count;
			LoadQuestion(m_SelectedIndex);
		}

		private void KeyPressed(object sender, KeyPressEventArgs e) {
			var ch = Char.ToUpperInvariant(e.KeyChar);
#if DEBUG
			string toDebug = String.Format("'{0}'", ch);
			if (!Char.IsLetterOrDigit(ch) && !Char.IsPunctuation(ch)) {
				toDebug = String.Format("0x{0:X2}", (int)(ch));
			}
			Debug.WriteLine(String.Format("DEBUG: KeyPress({0}, {1})", sender.GetType().Name, toDebug));
#endif
			switch (ch) {
				case 'A':
				case 'B':
				case 'C':
				case 'D':
					ProcessAnswer(ch);
					break;

				case 'P':
				case (char)(Keys.Left):
					PrevClick(sender, e);
					break;

				case 'N':
				case (char)(Keys.Right):
					NextClick(sender, e);
					break;

				case (char)(Keys.Return):
				case (char)(Keys.Space):
					if (!ReferenceEquals(imgResult.Image, null))
						NextClick(sender, e);
					break;

				case 'G':
					GoToSpecificQuestionClick(this, e);
					break;
			}
		}

		private void GoToFirstQuestionClicked(object sender, EventArgs e) {
			m_SelectedIndex = 0;
			LoadQuestion(0);
		}

		private void NavMouseDown(object sender, MouseEventArgs e) {
			if (imgResult.Image == null) return;
			var pos = PointToClient(MousePosition);
			var rect = new Rectangle(0, 0, Width, Height);
			if (rect.Contains(pos)) {
				m_Capture = true;
			} else {
				m_Capture = false;
			}
		}

		private void NavMouseUp(object sender, MouseEventArgs e) {
			if (this.m_Capture) {
				this.m_Capture = false;
				var pos = PointToClient(MousePosition);
				var rect = new Rectangle(0, 0, Width, Height);
				if (rect.Contains(pos)) {
					if (e.Button == System.Windows.Forms.MouseButtons.Left &&
						ReferenceEquals(imgResult.Image, m_Yes)) {
						NextClick(sender, e);
					}
				}
			}
		}

		private void ButtonEntered(object sender, EventArgs e) {
			dummyTarget.Select();
			if (ReferenceEquals(sender, tbQuestion) ||
				ReferenceEquals(sender, panelQuestion)) {
				if (ReferenceEquals(imgResult.Image, m_Yes)) {
					NextClick(sender, e);
				}
			}
		}

		private void StartCalculatorClick(object sender, EventArgs e) {
			System.Diagnostics.Process.Start("calc.exe");
		}

		private void RandomizeQuestionOrderClick(object sender, EventArgs e) {
			if (m_Questions.Count == 0) return;

			var working = new List<Question>(m_Questions.Values);
			m_Questions.Clear();
			var prng = new Random(DateTimeOffset.UtcNow.Millisecond);
			while (working.Count != 0) {
				var index = prng.Next(0, working.Count - 1);
				m_Questions.Add(working[index].Number, working[index]);
				working.RemoveAt(index);
			}
			LoadQuestion(0);
		}

		private void ShowQuestionsInOrderClick(object sender, EventArgs e) {
			if (m_Questions.Count == 0) return;

			// just reload the file
			if (File.Exists(m_LastFile))
				ProcessFile(m_LastFile);

			LoadQuestion(0);
		}

		private void GoToSpecificQuestionClick(object sender, EventArgs e) {
			using (var dialog = new GoToQuestionDialog()) {
				if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
					int seq = 0;
					if (Int32.TryParse(dialog.Question, out seq)) {
						if (seq >= 1 && seq <= m_Questions.Count) {
							LoadQuestion(seq - 1);
							return;
						}
					}
					seq = 0;
					foreach (var question in m_Questions) {
						if (question.Value.Number.Equals(dialog.Question, StringComparison.InvariantCultureIgnoreCase)) {
							LoadQuestion(seq);
							return;
						}
						seq++;
					}

				}
			}
		}

		private void ExitClick(object sender, EventArgs e) {
			this.Close();
		}

		private void AboutClick(object sender, EventArgs e) {
			MessageBox.Show(
			 String.Format("Radio Study GROL Version {0}\r\n\r\n", System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString()) +

			 "License: GNU General Public License Version 3.0.\r\n\r\n" +

			 "Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY. All rights reserved.\r\n\r\n" +

			 "This program is free software: you can redistribute it and/or modify " +
			 "it under the terms of the GNU General Public License as published by " +
			 "the Free Software Foundation, either version 3 of the License, or " +
			 "(at your option) any later version.\r\n\r\n" +

			 "This program is distributed in the hope that it will be useful, but " +
			 "WITHOUT ANY WARRANTY; without even the implied warranty of " +
			 "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU " +
			 "General Public License for more details.\r\n\r\n" +

			 "You should have received a copy of the GNU General Public License " +
			 "along with this program.  If not, see: http://www.gnu.org/licenses/", "About");
		}
		#endregion

		#region Utilities
		private void SortQuestionsBy(IList<string> keyOrder) {
			if (m_Questions.Count == 0) return;

			var working = new Dictionary<string, Question>(m_Questions);
			m_Questions.Clear();
			for (int i = 0; i != keyOrder.Count; ++i) {
				m_Questions.Add(keyOrder[i], working[keyOrder[i]]);
			}
		}

		/// <summary>
		/// Load a new question pool file.
		/// </summary>
		private bool ProcessFile(string file) {
			try {
				using (var ff = System.IO.File.OpenRead(file)) {
					using (var fs = new StreamReader(ff)) {
						var state = States.None;
						Question current = null;
						m_Questions.Clear();
						int lineNumber = 0;

						while (!fs.EndOfStream) {
							var line = fs.ReadLine();
							lineNumber++;
							if (ReferenceEquals(line, null)) continue;
							line = line.Trim();

							// empty line?
							if (String.IsNullOrEmpty(line) || String.IsNullOrWhiteSpace(line) || line.Equals("~~")) {
								state = States.None;
								if (!ReferenceEquals(current, null)) {
									if (current.Answers.Count == 4) {
										m_Questions[current.Number] = current;
										current = null;
									} else if (current.Answers.Count != 0) {
										MessageBox.Show("Question " + current.Number + " has an insufficient number of answers.", "Error");
										m_Questions.Clear();
										tbQuestion.Clear();
										return false;
									}
								}
								continue;
							}

							// answer key footer?
							if (line.StartsWith("Answer Key", StringComparison.InvariantCultureIgnoreCase) ||
								line.StartsWith("Answers:", StringComparison.InvariantCultureIgnoreCase)) {

								if (!ReferenceEquals(current, null)) {
									MessageBox.Show("Question " + current.Number + " on line " + lineNumber.ToString() + " left dangling.", "Error");
									m_Questions.Clear();
									tbQuestion.Clear();
									return false;
								}

								var parts = line.Split(new[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
								var key = "";
								foreach (var item in parts) {
									if (IsAnswerToken(item)) {
										key = item;
										if (key.EndsWith(":")) {
											key = key.Substring(0, key.Length - 1);
											if (!m_Questions.ContainsKey(key)) {
												MessageBox.Show("Answer " + key + " on line " + lineNumber.ToString() + " is a dangling answer.", "Error");
												m_Questions.Clear();
												tbQuestion.Clear();
												return false;
											}
										}
									} else if (item.Length == 1 && item[0] >= 'A' && item[0] <= 'D') {
										if (!m_Questions.ContainsKey(key)) {
											MessageBox.Show("Answer " + key + " on line " + lineNumber.ToString() + " is a dangling answer.", "Error");
											m_Questions.Clear();
											tbQuestion.Clear();
											return false;
										}
										m_Questions[key].Letter = item[0];
									}
								}

							// new question?
							} else if (IsNewQuestion(line)) {
								if (!ReferenceEquals(current, null)) {
									MessageBox.Show("Question " + current.Number + " on line " + lineNumber.ToString() + " left dangling.", "Error");
									m_Questions.Clear();
									tbQuestion.Clear();
									return false;
								}

								var idx = line.IndexOfAny(new[] { ' ', '\t' });

								state = States.Question;
								if (idx > 0) {
									var text = line.Substring(idx).Trim();
									if (text.StartsWith("(A)", StringComparison.InvariantCultureIgnoreCase) ||
										text.StartsWith("(B)", StringComparison.InvariantCultureIgnoreCase) ||
										text.StartsWith("(C)", StringComparison.InvariantCultureIgnoreCase) ||
										text.StartsWith("(D)", StringComparison.InvariantCultureIgnoreCase)) {
										current = new Question {
											Number = line.Substring(0, idx),
											Text = "",
											Letter = text[1],
										};
									} else {
										current = new Question {
											Number = line.Substring(0, idx),
											Text = text
										};
									}
								} else {
									current = new Question {
										Number = line,
										Text = ""
									};
								}
							} else if (line.Length == 1 && line[0] >= 'A' && line[0] <= 'D') {
								if (!ReferenceEquals(current, null)) {
									current.Letter = line[0];
								}
							} else if (line.Length >= 2 && line[1] == '.' && line[0] >= 'A' && line[0] <= 'D') {
								if (ReferenceEquals(current, null)) {
									MessageBox.Show("Orphaned answer found on line " + lineNumber.ToString(), "Error");
									m_Questions.Clear();
									tbQuestion.Clear();
									return false;
								}

								state = States.Answer;
								var ans = new Answer { Letter = line[0], Text = line.Substring(2).Trim() };
								current.Answers.Add(ans);
							} else {
								switch (state) {
									case States.Question: {
										if (current.Text.Length != 0)
											current.Text += ' ';
										current.Text += line;
									} break;
									case States.Answer: {
											if (current.Answers.Count != 0) {
												current.Answers.Last().Text += (' ' + line);
											}
									} break;
								}
							}
						}
					}
				}

				m_SelectedIndex = 0;
				LoadQuestion(m_SelectedIndex);
				m_LastPath = System.IO.Path.GetDirectoryName(file);
				LastFile = file;
				return true;
			} catch (Exception ex) {
				MessageBox.Show(String.Format("Could not open file: {0}", ex.Message), "Error");
				return false;
			}
		}

		private bool IsAnswerToken(string item) {
			if (Char.IsDigit(item[0]) && item[1] == '-') return true;
			if (Regex.IsMatch(item, "^[0-9]{1,3}[A-Z][0-9]{1,3}[:]{0,1}$")) return true;
			return false;
		}

		private bool IsNewQuestion(string line) {
			if (line.Length < 2) return false;
			if (Char.IsDigit(line[0]) && line[1] == '-' && Char.IsDigit(line[2])) return true;
			if (Regex.IsMatch(line, "^[0-9]{1,3}[A-Z][0-9]{1,3} ")) return true;
			if (Regex.IsMatch(line, "^6A[0-9]{1,3}$")) return true;
			if (Regex.IsMatch(line, "^[EGT][0-9][A-Z][0-9]{2}")) return true;
			return false;
		}

		/// <summary>
		/// Check the given answer, and respond accordingly.
		/// </summary>
		private void ProcessAnswer(char answer) {
			if (m_SelectedQuestion == null) return;
			if (m_SelectedQuestion.Letter == answer) {
				imgResult.Image = m_Yes;
				foreach (var button in new[] { bA, bB, bC, bD }) {
					if (button.Tag.Equals(answer)) {
						button.BackColor = Color.Lime;
					}
				}
			} else {
				imgResult.Image = m_No;
			}
		}

		/// <summary>
		/// Display a new question.
		/// </summary>
		private void LoadQuestion(int index) {
			m_SelectedIndex = index;
			foreach (var button in new[] { bA, bB, bC, bD }) {
				button.BackColor = SystemColors.Control;
			}
			if (index < 0 || m_Questions.Count == 0) {
				tbQuestion.Text = "";
				m_SelectedQuestion = null;
			} else {
				var keys = new List<string>(m_Questions.Keys);
				//keys.Sort();
				var key = keys[index];
				m_SelectedQuestion = m_Questions[key];
				tbQuestion.Text = String.Format(
					"{0} ({1} of {2} - {3}%)\r\n\r\n{4}\r\n\r\n",
					m_SelectedQuestion.Number,
					index + 1,
					m_Questions.Count,
					(100 * (index + 1)) / m_Questions.Count,
					m_SelectedQuestion.Text);
				for (int i = 0; i != m_SelectedQuestion.Answers.Count; ++i) {
					tbQuestion.Text += ("      " + m_SelectedQuestion.Answers[i].Letter + ".  " + m_SelectedQuestion.Answers[i].Text + "\r\n");
				}
			}
			imgResult.Image = null;
		}

		/// <summary>
		/// Save the current program state.
		/// </summary>
		private void SaveConfig() {
			try {
				if (!ReferenceEquals(LastFile, null)) {
					if (!Directory.Exists(ConfigPath)) {
						Directory.CreateDirectory(ConfigPath);
					}
					XmlDocument doc = new XmlDocument();
					doc.AppendChild(doc.CreateElement("Config"));
					doc.DocumentElement.SetAttribute("LastFile", LastFile);
					doc.DocumentElement.SetAttribute("LastPath", m_LastPath);
					doc.DocumentElement.SetAttribute("LastIndex", m_SelectedIndex.ToString());
					doc.DocumentElement.SetAttribute("X", Location.X.ToString());
					doc.DocumentElement.SetAttribute("Y", Location.Y.ToString());
					doc.DocumentElement.SetAttribute("Width", Width.ToString());
					doc.DocumentElement.SetAttribute("Height", Height.ToString());
					doc.DocumentElement.SetAttribute("State", WindowState.ToString());
					if (m_Questions.Count != 0) {
						var qo = doc.CreateElement("QuestionOrder");
						doc.DocumentElement.AppendChild(qo);
						foreach (var key in m_Questions.Keys) {
							var q = doc.CreateElement("Question");
							q.InnerText = key;
							qo.AppendChild(q);
						}
					}
					doc.Save(ConfigFile);
				}
			} catch {
				// nop
			}
		}

		/// <summary>
		/// Load the prior program state.
		/// </summary>
		private void LoadConfig() {
			try {
				if (System.IO.File.Exists(ConfigFile)) {
					XmlDocument doc = new XmlDocument();
					doc.Load(ConfigFile);
					LastFile = doc.DocumentElement.GetAttribute("LastFile");
					m_LastPath = doc.DocumentElement.GetAttribute("LastPath");

					if (doc.DocumentElement.HasAttribute("State")) {
						WindowState = (FormWindowState)(Enum.Parse(typeof(FormWindowState), doc.DocumentElement.GetAttribute("State")));
					}
					if (doc.DocumentElement.HasAttribute("X") && doc.DocumentElement.HasAttribute("Y")) {
						int x = Int32.Parse(doc.DocumentElement.GetAttribute("X"));
						int y = Int32.Parse(doc.DocumentElement.GetAttribute("Y"));
						Location = new Point(x, y);
					}
					if (doc.DocumentElement.HasAttribute("Width") && doc.DocumentElement.HasAttribute("Height") && WindowState != FormWindowState.Maximized) {
						int w = Int32.Parse(doc.DocumentElement.GetAttribute("Width"));
						int h = Int32.Parse(doc.DocumentElement.GetAttribute("Height"));
						Size = new Size(w, h);
					}

					ProcessFile(LastFile);

					m_SelectedIndex = Int32.Parse(doc.DocumentElement.GetAttribute("LastIndex"));

					foreach (var e in doc.DocumentElement.ChildNodes.OfType<XmlElement>()) {
						switch (e.LocalName) {
							case "QuestionOrder":
								var keyOrder = new List<string>();
								foreach (var f in e.ChildNodes.OfType<XmlElement>()) {
									if (String.IsNullOrEmpty(f.InnerText)) break;
									var key = f.InnerText.Trim();
									if (!m_Questions.ContainsKey(key)) {
										break;
									}
									keyOrder.Add(key);
								}
								SortQuestionsBy(keyOrder);
								break;
						}
					}

					// load the last question
					LoadQuestion(m_SelectedIndex);
				}
			} catch {
				// nop
			}
		}

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main() {
			Application.EnableVisualStyles();
			Application.SetCompatibleTextRenderingDefault(false);
			Application.Run(new MainForm());
		}
		#endregion
	}
}
